import math
import numpy as np

from typing import Union

from .data_types import *
from .distributions import *
from ..ml import Grid
from ..misc import prod


class ParamsGenerator:
    """
    Parameter generator for param searching, see ./dist/bases.py -> class SearchBase for usage

    Parameters
    ----------
    params : {DataType, Iterable, dict}, parameter settings
        * DataType: distribution of this DataType must be `Choice`. In this case, we'll simply 'enumerate'
        through the config choices
        * Iterable: each element should be a DataType. It could be nested
        * dict: each key should correspond to a config key, and its value should correspond to the
        desired value distribution. It could be nested

    Examples
    ----------
    >>> grid = ParamsGenerator({
    >>>     "a": Any(Choice(values=[1, 2, 3])),
    >>>     "c": {"d": Int(Choice(values=[1, 2, 3])), "e": Float(Choice(values=[1, 2]))}
    >>> })
    >>> for param in grid.all():
    >>>     print(param)
    >>> # output : {'a': 1, 'c': {'d': 1, 'e': 1, 'f': 3}}, {'a': 1, 'c': {'d': 1, 'e': 1, 'f': 4}}
    >>> #          {'a': 1, 'c': {'d': 1, 'e': 2, 'f': 3}}, {'a': 1, 'c': {'d': 1, 'e': 2, 'f': 4}}
    >>> #          {'a': 1, 'c': {'d': 2, 'e': 1, 'f': 3}}, {'a': 1, 'c': {'d': 2, 'e': 1, 'f': 4}}
    >>> #          {'a': 1, 'c': {'d': 2, 'e': 2, 'f': 3}}, {'a': 1, 'c': {'d': 2, 'e': 2, 'f': 4}}
    >>> #          ......
    >>> #          {'a': 3, 'c': {'d': 3, 'e': 2, 'f': 3}}, {'a': 3, 'c': {'d': 3, 'e': 2, 'f': 4}}

    """

    def __init__(self, params):
        self._params = params
        self._delim, self._idx_start = "^_^", "$$$"
        if isinstance(self._params, DataType):
            assert_msg = "distribution must be `Choice` when DataType is used as `params`"
            assert isinstance(self._params._distribution, Choice), assert_msg

    @property
    def n_params(self):
        if self.is_enumerate:
            return self._params.n_params
        def _n_params(params):
            if isinstance(params, (DataType, Iterable)):
                return params.n_params
            assert isinstance(params, dict)
            n_params = prod(_n_params(v) for v in params.values())
            if math.isinf(n_params):
                return n_params
            return int(n_params)
        return _n_params(self._params)

    @property
    def is_enumerate(self):
        return isinstance(self._params, (DataType, Iterable))

    @staticmethod
    def _append_ax_param(name: str, param: DataType, flattened_params):
        param_dist = param._distribution
        if isinstance(param_dist, Choice):
            local_params = param.all()
            if len(local_params) == 1:
                flattened_params.append({"name": name, "type": "fixed", "value": local_params[0]})
            else:
                flattened_params.append({"name": name, "type": "choice", "values": local_params})
        else:
            log_scale = isinstance(param_dist, Exponential)
            if isinstance(param, Int):
                param_type = "int"
            elif isinstance(param, Float):
                param_type = "float"
            elif isinstance(param, Bool):
                param_type = "bool"
            elif isinstance(param, String):
                param_type = "str"
            else:
                raise NotImplementedError
            flattened_params.append({
                "name": name, "type": "range", "bounds": [param.lower, param.upper],
                "parameter_type": param_type, "log_scale": log_scale
            })

    def pop(self):
        if self.is_enumerate:
            return self._params.pop()
        def _pop(src: dict, tgt: dict):
            for k, v in src.items():
                if isinstance(v, dict):
                    next_tgt = tgt.setdefault(k, {})
                    _pop(v, next_tgt)
                else:
                    tgt[k] = v.pop()
            return tgt
        return _pop(self._params, {})

    def all(self):
        if self.is_enumerate:
            yield from self._params.all()
        else:
            def _all(src: dict, tgt: dict):
                for k, v in src.items():
                    if isinstance(v, dict):
                        next_tgt = tgt.setdefault(k, {})
                        _all(v, next_tgt)
                    else:
                        tgt[k] = v.all()
                return tgt
            all_flattened_params_list = self.get_flattened_params(_all(self._params, {}))
            all_flattened_params = dict(all_flattened_params_list)
            for flattened_params in Grid(all_flattened_params):
                yield self.nest_flattened_params(flattened_params)

    def get_flattened_params(self, tgt_params=None, use_ax=False):
        if tgt_params is None:
            tgt_params = self._params
        if isinstance(tgt_params, (DataType, Iterable)):
            raise ValueError("get_flattened_params method requires nested params")
        flattened_params = []
        def _flatten_params(d: dict, pre_key: Union[None, str]):
            for name, params in d.items():
                if pre_key is None:
                    next_pre_key = name
                else:
                    next_pre_key = f"{pre_key}{self._delim}{name}"
                if isinstance(params, dict):
                    _flatten_params(params, next_pre_key)
                elif isinstance(params, Iterable):
                    for i, param in enumerate(params._values):
                        name = f"{next_pre_key}{self._delim}{self._idx_start}{i}"
                        if not use_ax:
                            flattened_params.append((name, param))
                        else:
                            self._append_ax_param(name, param, flattened_params)
                else:
                    if not use_ax:
                        flattened_params.append((next_pre_key, params))
                    else:
                        self._append_ax_param(next_pre_key, params, flattened_params)
            return flattened_params
        return _flatten_params(tgt_params, None)

    def nest_flattened_params(self, flattened_params):
        sorted_params = sorted(map(
            lambda k, v: (k.split(self._delim), v),
            *zip(*flattened_params.items())
        ), key=len)
        l_start = len(self._idx_start)
        list_traces, nested_params = {}, {}
        for key_list, value in sorted_params:
            if len(key_list) == 1:
                nested_params[key_list[0]] = value
            else:
                last_key = key_list[-1]
                if last_key.startswith(self._idx_start):
                    list_traces.setdefault(tuple(key_list[:-1]), []).append((int(last_key[l_start:]), value))
                else:
                    parent = nested_params.setdefault(key_list[0], {})
                    for key in key_list[1:-1]:
                        parent = parent.setdefault(key, {})
                    parent[last_key] = value
        for list_key_tuple, list_values in list_traces.items():
            if len(list_key_tuple) == 1:
                parent = nested_params
            else:
                parent = nested_params.setdefault(list_key_tuple[0], {})
                for key in list_key_tuple[1:-1]:
                    parent = parent.setdefault(key, {})
            indices, values = zip(*list_values)
            assert sorted(indices) == list(range(len(indices)))
            d = self._params
            for key in list_key_tuple[:-1]:
                d = d[key]
            constructor = d[list_key_tuple[-1]]._constructor
            parent[list_key_tuple[-1]] = constructor(values[i] for i in np.argsort(indices))
        return nested_params


__all__ = ["ParamsGenerator"]



from .basis import (
    ParamsGenerator
)
from .data_types import (
    DataType,
    Iterable,
    Any,
    Int,
    Float,
    Bool,
    String
)
from .distributions import (
    DistributionBase,
    Uniform,
    Exponential,
    Choice
)


__all__ = [
    "ParamsGenerator",
    "DataType", "Iterable", "Any", "Int", "Float", "Bool", "String",
    "DistributionBase", "Uniform", "Exponential", "Choice"
]



import math
import random

from abc import *
from typing import List, Tuple, Union

from ..ml import Grid
from ..misc import prod
from .distributions import DistributionBase


class DataType(metaclass=ABCMeta):
    def __init__(self, distribution: DistributionBase = None, optional=True, **kwargs):
        self._optional, self._distribution, self.config = optional, distribution, kwargs

    @property
    @abstractmethod
    def n_params(self):
        raise NotImplementedError

    @abstractmethod
    def transform(self, value):
        raise NotImplementedError

    def __str__(self):
        return f"{type(self).__name__}({self._distribution})"

    __repr__ = __str__

    @property
    def distribution_is_inf(self):
        return math.isinf(self._distribution.n_params)

    def _all(self):
        return list(map(self.transform, self._distribution._values))

    def pop(self):
        return self.transform(self._distribution.pop())

    def all(self):
        if math.isinf(self.n_params):
            raise ValueError("'all' method could be called iff n_params is finite")
        return self._all()

    @property
    def lower(self):
        dist_lower = self._distribution._lower
        if dist_lower is None:
            return
        return self.transform(dist_lower)

    @property
    def upper(self):
        dist_upper = self._distribution._upper
        if dist_upper is None:
            return
        return self.transform(dist_upper)

    @property
    def values(self):
        dist_values = self._distribution._values
        if dist_values is None:
            return
        return list(map(self.transform, dist_values))


class Iterable:
    def __init__(self, values: Union[List[DataType], Tuple[DataType]]):
        self._values = values
        self._constructor = list if isinstance(values, list) else tuple

    def __str__(self):
        braces = "[]" if self._constructor is list else "()"
        return f"{braces[0]}{', '.join(map(str, self._values))}{braces[1]}"

    __repr__ = __str__

    def pop(self):
        return self._constructor(v.pop() for v in self._values)

    def all(self, return_values=False):
        grid = Grid([v.all() for v in self._values])
        generator = (self._constructor(v) for v in grid)
        if return_values:
            return list(generator)
        yield from generator

    @property
    def values(self):
        return self._values

    @property
    def n_params(self):
        n_params = prod(v.n_params for v in self._values)
        if math.isinf(n_params):
            return n_params
        return int(n_params)


class Any(DataType):
    @property
    def n_params(self):
        return self._distribution.n_params

    def transform(self, value):
        return value


class Int(DataType):
    @property
    def lower(self):
        return int(math.ceil(self._distribution._lower))

    @property
    def upper(self):
        return int(math.floor(self._distribution._upper))

    @property
    def values(self):
        return list(range(self.lower, self.upper + 1))

    @property
    def n_params(self):
        if self.distribution_is_inf:
            return int(self.upper - self.lower) + 1
        return self._distribution.n_params

    def _all(self):
        if self.distribution_is_inf:
            return list(range(self.lower, self.upper + 1))
        return super()._all()

    def transform(self, value):
        return int(round(value + random.random() * 2e-4 - 1e-4))


class Float(DataType):
    @property
    def n_params(self):
        return self._distribution.n_params

    def transform(self, value):
        return float(value)


class Bool(DataType):
    @property
    def n_params(self):
        if self.distribution_is_inf:
            return 2
        return len(self._all())

    def _all(self):
        return sorted(super()._all())

    def transform(self, value):
        return bool(value)


class String(DataType):
    @property
    def n_params(self):
        return self._distribution.n_params

    def transform(self, value):
        return str(value)


__all__ = ["DataType", "Iterable", "Any", "Int", "Float", "Bool", "String"]



import math
import random

from abc import *


class DistributionBase(metaclass=ABCMeta):
    def __init__(self, lower=None, upper=None, values=None, **kwargs):
        self._lower, self._upper, self._values, self.config = lower, upper, values, kwargs

    @property
    @abstractmethod
    def n_params(self):
        raise NotImplementedError

    @abstractmethod
    def pop(self):
        raise NotImplementedError

    def __str__(self):
        if self._values is not None:
            return f"{type(self).__name__}[{', '.join(map(str, self._values))}]"
        return f"{type(self).__name__}[{self._lower}, {self._upper}]"

    __repr__ = __str__

    def _assert_lower_and_upper(self):
        assert self._lower is not None, "lower should be provided"
        assert self._upper is not None, "upper should be provided"

    def _assert_values(self):
        assert isinstance(self._values, list), "values should be a list"


class Uniform(DistributionBase):
    @property
    def n_params(self):
        return math.inf

    def pop(self):
        self._assert_lower_and_upper()
        return random.random() * (self._upper - self._lower) + self._lower


class Exponential(DistributionBase):
    @property
    def n_params(self):
        return math.inf

    def pop(self):
        self._assert_lower_and_upper()
        assert self._lower > 0, "lower should be greater than 0 in exponential distribution"
        ratio = self.config.setdefault("ratio", 2)
        assert ratio > 1, "ratio should be greater than 1"
        diff = math.log(self._upper / self._lower, ratio)
        return math.pow(ratio, random.random() * diff) * self._lower


class Choice(DistributionBase):
    @property
    def n_params(self):
        return len(self._values)

    def pop(self):
        self._assert_values()
        return random.choice(self._values)


__all__ = ["DistributionBase", "Uniform", "Exponential", "Choice"]



import io
import os
import sys
import dill
import json
import math
import time
import errno
import random
import shutil
import inspect
import logging
import hashlib
import zipfile
import datetime
import operator
import threading
import unicodedata

import numpy as np
import matplotlib.pyplot as plt

from typing import *
from PIL import Image
from functools import reduce
from collections import Counter
from abc import abstractmethod

dill._dill._reverse_typemap["ClassType"] = type


# util functions

def timestamp(simplify: bool = False,
              ensure_different: bool = False) -> str:
    """
    Return current timestamp

    Parameters
    ----------
    simplify : bool. If True, format will be simplified to 'year-month-day'
    ensure_different : bool. If True, format will include millisecond

    Returns
    -------
    timestamp : str

    """

    now = datetime.datetime.now()
    if simplify:
        return now.strftime("%Y-%m-%d")
    if ensure_different:
        return now.strftime("%Y-%m-%d_%H-%M-%S-%f")
    return now.strftime("%Y-%m-%d_%H-%M-%S")


def prod(iterable: Iterable) -> float:
    """ return cumulative production of an iterable """

    return float(reduce(operator.mul, iterable, 1))


def hash_code(code: str) -> str:
    """ return hash code for a string """

    code = code.encode()
    return hashlib.md5(code).hexdigest()[:8]


def prefix_dict(d: Dict[str, Any], prefix: str):
    """ prefix every key in dict `d` with `prefix` """

    return {f"{prefix}_{k}": v for k, v in d.items()}


def shallow_copy_dict(d: dict) -> dict:
    d = d.copy()
    for k, v in d.items():
        if isinstance(v, dict):
            d[k] = shallow_copy_dict(v)
    return d


def update_dict(src_dict: dict, tgt_dict: dict) -> dict:
    """
    Update tgt_dict with src_dict
    * Notice that changes will happen only on keys which src_dict holds

    Parameters
    ----------
    src_dict : dict
    tgt_dict : dict

    Returns
    -------
    tgt_dict : dict

    """

    for k, v in src_dict.items():
        tgt_v = tgt_dict.get(k)
        if tgt_v is None:
            tgt_dict[k] = v
        elif not isinstance(v, dict):
            tgt_dict[k] = v
        else:
            update_dict(v, tgt_v)
    return tgt_dict


def fix_float_to_length(num: float, length: int) -> str:
    """ change a float number to string format with fixed length """

    str_num = f"{num:f}"
    if str_num == "nan":
        return f"{str_num:^{length}s}"
    length = max(length, str_num.find("."))
    return str_num[:length].ljust(length, "0")


def truncate_string_to_length(string: str, length: int) -> str:
    """ truncate a string to make sure its length not exceeding a given length """

    if len(string) <= length:
        return string
    half_length = int(0.5 * length) - 1
    return string[:half_length] + "." * (length - 2 * half_length) + string[-half_length:]


def grouped(iterable: Iterable, n: int, *, keep_tail=False) -> List[tuple]:
    """ group an iterable every `n` elements """

    if not keep_tail:
        return list(zip(*[iter(iterable)] * n))
    with batch_manager(object(), iterable, batch_size=n) as manager:
        return [tuple(batch) for batch in manager]


def is_numeric(s: str) -> bool:
    """ check whether `s` is a number """

    try:
        s = float(s)
        return True
    except (TypeError, ValueError):
        try:
            unicodedata.numeric(s)
            return True
        except (TypeError, ValueError):
            return False


def get_one_hot(feature: Union[list, np.ndarray], dim: int) -> np.ndarray:
    """
    Get one-hot representation

    Parameters
    ----------
    feature : array-like, source data of one-hot representation
    dim : int, dimension of the one-hot representation

    Returns
    -------
    one_hot : np.ndarray, one-hot representation of `feature`

    """

    one_hot = np.zeros([len(feature), dim], np.int64)
    one_hot[range(len(one_hot)), np.asarray(feature, np.int64).ravel()] = 1
    return one_hot


def show_or_save(export_path: str,
                 fig: plt.figure = None, **kwargs) -> None:
    """
    Utility function to deal with figure

    Parameters
    ----------
    export_path : {None, str}
        * None : the figure will be shown
        * str  : it represents the path where the figure should be saved to
    fig : {None, plt.Figure}
        * None       : default figure contained in plt will be executed
        * plt.Figure : it will be executed

    """

    if export_path is None:
        fig.show(**kwargs) if fig is not None else plt.show(**kwargs)
    else:
        fig.savefig(export_path) if fig is not None else plt.savefig(export_path, **kwargs)
    plt.close()


def show_or_return(return_canvas: bool) -> Union[None, np.ndarray]:
    """
    Utility function to deal with current plt

    Parameters
    ----------
    return_canvas : bool, whether return canvas or not

    """

    if not return_canvas:
        plt.show()
        return

    buffer_ = io.BytesIO()
    plt.savefig(buffer_, format="png")
    plt.close()
    buffer_.seek(0)
    image = Image.open(buffer_)
    canvas = np.asarray(image)[..., :3]
    buffer_.close()
    return canvas


def get_indices_from_another(base: np.ndarray,
                             segment: np.ndarray) -> np.ndarray:
    """
    Get `segment` elements' indices in `base`

    Warnings
    ----------
    All elements in segment should appear in base to ensure validity

    Parameters
    ----------
    base : np.ndarray, base array
    segment : np.ndarray, segment array

    Returns
    -------
    indices : np.ndarray, positions where elements in `segment` appear in `base`

    Examples
    -------
    >>> import numpy as np
    >>> base, segment = np.arange(100), np.random.permutation(100)[:10]
    >>> assert np.allclose(get_indices_from_another(base, segment), segment)

    """
    base_sorted_args = np.argsort(base)
    positions = np.searchsorted(base[base_sorted_args], segment)
    return base_sorted_args[positions]


class UniqueIndices(NamedTuple):
    """
    unique           : np.ndarray, unique values of the given array (`arr`)
    unique_cnt       : np.ndarray, counts of each unique value
    sorting_indices  : np.ndarray, indices which can (stably) sort the given array by its value
    split_arr        : np.ndarray, array which can split the `sorting_indices` to make sure that
                       each portion of the split indices belong & only belong to one of the unique values
    """
    unique: np.ndarray
    unique_cnt: np.ndarray
    sorting_indices: np.ndarray
    split_arr: np.ndarray

    @property
    def split_indices(self):
        return np.split(self.sorting_indices, self.split_arr)


def get_unique_indices(arr: np.ndarray) -> UniqueIndices:
    """
    Get indices for unique values of an array

    Parameters
    ----------
    arr : np.ndarray, target array which we wish to find indices of each unique value

    Returns
    -------
    UniqueIndices

    Examples
    -------
    >>> import numpy as np
    >>> arr = np.array([1, 2, 3, 2, 4, 1, 0, 1], np.int64)
    >>> print(get_unique_indices(arr, return_raw=True), get_unique_indices(arr)[-1])
    >>> # [0, 1, 2, 3, 4]
    >>> # [1, 3, 2, 1, 1]
    >>> # [6, 0, 5, 7, 1, 3, 2, 4]
    >>> # [1, 4, 6, 7]
    >>> # [ [6], [0, 5, 7], [1, 3], [2], [4] ]

    """
    unique, unique_inv, unique_cnt = np.unique(arr, return_inverse=True, return_counts=True)
    sorting_indices, split_arr = np.argsort(unique_inv, kind="mergesort"), np.cumsum(unique_cnt)[:-1]
    return UniqueIndices(unique, unique_cnt, sorting_indices, split_arr)


def get_counter_from_arr(arr):
    if isinstance(arr, np.ndarray):
        arr = dict(zip(*np.unique(arr, return_counts=True)))
    return Counter(arr)


def register_core(name: str,
                  global_dict: Dict[str, type], *,
                  before_register: callable = None,
                  after_register: callable = None):
    def _register(cls):
        if before_register is not None:
            before_register(cls)
        registered = global_dict.get(name)
        if registered is not None:
            print(f"~~~ [warning] '{name}' has already registered "
                  f"in the given global dict ({global_dict})")
            return cls
        global_dict[name] = cls
        if after_register is not None:
            after_register(cls)
        return cls
    return _register


# util modules

class Incrementer:
    """
    Util class which can calculate running mean & running std efficiently

    Parameters
    ----------
    window_size : {int, None}, window size of running statistics
        * None : then all history records will be used for calculation

    Examples
    ----------
    >>> incrementer = Incrementer(window_size=5)
    >>> for i in range(10):
    >>>     incrementer.update(i)
    >>>     if i >= 4:
    >>>         print(incrementer.mean)  # will print 2.0, 3.0, ..., 6.0, 7.0

    """

    def __init__(self, window_size: int = None):
        if window_size is not None:
            if not isinstance(window_size, int):
                raise ValueError(f"window size should be integer, {type(window_size)} found")
            if window_size < 2:
                raise ValueError(f"window size should be greater than 2, {window_size} found")
        self._window_size = window_size
        self._n_record = self._running_sum = self._running_square_sum = self._previous = None

    @property
    def mean(self):
        return self._running_sum / self._n_record

    @property
    def std(self):
        return math.sqrt(max(self._running_square_sum / self._n_record - self.mean ** 2, 0.))

    @property
    def n_record(self):
        return self._n_record

    def update(self, new_value):
        if self._n_record is None:
            self._n_record, self._running_sum, self._running_square_sum = 1, new_value, new_value ** 2
        else:
            self._n_record += 1
            self._running_sum += new_value
            self._running_square_sum += new_value ** 2
        if self._window_size is not None:
            if self._previous is None:
                self._previous = [new_value]
            else:
                self._previous.append(new_value)
            if self._n_record == self._window_size + 1:
                self._n_record -= 1
                previous = self._previous.pop(0)
                self._running_sum -= previous
                self._running_square_sum -= previous ** 2


class _Formatter(logging.Formatter):
    """ formatter for logging, which supports millisecond """

    converter = datetime.datetime.fromtimestamp

    def formatTime(self, record, datefmt=None):
        ct = self.converter(record.created)
        if datefmt:
            s = ct.strftime(datefmt)
        else:
            t = ct.strftime("%Y-%m-%d %H:%M:%S")
            s = "%s.%03d" % (t, record.msecs)
        return s


class LoggingMixin:
    """
    Mixin class to provide logging methods for base class

    Attributes
    ----------
    _triggered_ : bool
        If not `_triggered_`, log file will not be created

    _verbose_level_ : int
        Preset verbose level of the whole logging process

    Methods
    ----------
    log_msg(self, body, prefix="", verbose_level=1)
        Log something either through console or to a file
        * body : str
            main logging message
        * prefix : str
            prefix added to `body` when logging message goes through console
        * verbose_level : int
            if `self._verbose_level_` >= verbose_level, then logging message will go through console

    log_block_msg(self, body, prefix="", title="", verbose_level=1)
        Almost the same as `log_msg`, except adding `title` on top of `body`

    """

    _triggered_ = False
    _initialized_ = False
    _logging_path_ = None
    _logger_ = _verbose_level_ = None
    _date_format_string_ = "%Y-%m-%d %H:%M:%S.%f"
    _formatter_ = _Formatter(
        "[ {asctime:s} ] [ {levelname:^8s} ] {func_prefix:s} {message:s}",
        _date_format_string_, style="{"
    )
    _timing_dict_, _time_cache_dict_ = {}, {}

    info_prefix = "~~~  [ info ] "
    warning_prefix = "~~~ [warning] "
    error_prefix = "~~~ [ error ] "

    @property
    def logging_path(self):
        if self._logging_path_ is None:
            folder = os.path.join(os.getcwd(), "_logging", type(self).__name__)
            os.makedirs(folder, exist_ok=True)
            self._logging_path_ = os.path.join(folder, f"{timestamp()}.log")
        return self._logging_path_

    @property
    def console_handler(self):
        if self._logger_ is None:
            return
        for handler in self._logger_.handlers:
            if isinstance(handler, logging.StreamHandler):
                return handler

    @staticmethod
    def _get_func_prefix(frame=None, return_prefix=True):
        if frame is None:
            frame = inspect.currentframe().f_back.f_back
        if not return_prefix:
            return frame
        frame_info = inspect.getframeinfo(frame)
        file_name = truncate_string_to_length(os.path.basename(frame_info.filename), 16)
        func_name = truncate_string_to_length(frame_info.function, 24)
        func_prefix = f"[ {func_name:^24s} ] [ {file_name:>16s}:{frame_info.lineno:<4d} ]"
        return func_prefix

    @staticmethod
    def _release_handlers(logger):
        for handler in logger.handlers[:]:
            handler.close()
            logger.removeHandler(handler)

    def _init_logging(self,
                      verbose_level: Union[int, None] = 2,
                      trigger: bool = True):
        if LoggingMixin._initialized_:
            return self
        LoggingMixin._initialized_ = True
        logger_name = getattr(self, "_logger_name_", "root")
        logger = LoggingMixin._logger_ = logging.getLogger(logger_name)
        LoggingMixin._verbose_level_ = verbose_level
        if not trigger:
            return self
        LoggingMixin._triggered_ = True
        config = getattr(self, "config", {})
        self._logging_path_ = config.get("_logging_path_")
        if self._logging_path_ is None:
            self._logging_path_ = config["_logging_path_"] = self.logging_path
        os.makedirs(os.path.dirname(self.logging_path), exist_ok=True)
        file_handler = logging.FileHandler(self.logging_path)
        file_handler.setFormatter(self._formatter_)
        file_handler.setLevel(logging.DEBUG)
        console = logging.StreamHandler()
        console.setLevel(logging.INFO)
        console.setFormatter(_Formatter("{custom_prefix:s}{message:s}", style="{"))
        logger.setLevel(logging.DEBUG)
        self._release_handlers(logger)
        logger.addHandler(console)
        logger.addHandler(file_handler)
        for handler in logging.getLogger().handlers:
            handler.setLevel(logging.CRITICAL)
        self.log_block_msg(sys.version, title="system version", verbose_level=None)
        return self

    def log_msg(self,
                body: str,
                prefix: str = "",
                verbose_level: Union[int, None] = 1,
                msg_level: int = logging.INFO,
                frame=None):
        preset_verbose_level = getattr(self, "_verbose_level", None)
        if preset_verbose_level is not None:
            self._verbose_level_ = preset_verbose_level
        elif self._verbose_level_ is None:
            self._verbose_level_ = 0
        console_handler = self.console_handler
        if verbose_level is None or self._verbose_level_ < verbose_level:
            do_print, console_level = False, msg_level + 10
        else:
            do_print, console_level = not LoggingMixin._triggered_, msg_level
        if console_handler is not None:
            console_handler.setLevel(console_level)
        if do_print:
            print(prefix + body)
        elif LoggingMixin._triggered_:
            func_prefix = self._get_func_prefix(frame)
            self._logger_.log(msg_level, body, extra={"func_prefix": func_prefix, "custom_prefix": prefix})
        if console_handler is not None:
            console_handler.setLevel(logging.INFO)

    def log_block_msg(self,
                      body: str,
                      prefix: str = "",
                      title: str = "",
                      verbose_level: Union[int, None] = 1,
                      msg_level: int = logging.INFO,
                      frame=None):
        frame = self._get_func_prefix(frame, False)
        self.log_msg(f"{title}\n{body}\n", prefix, verbose_level, msg_level, frame)

    def exception(self, body, frame=None):
        self._logger_.exception(body, extra={
            "custom_prefix": self.error_prefix,
            "func_prefix": LoggingMixin._get_func_prefix(frame)
        })

    @staticmethod
    def log_with_external_method(body, prefix, log_method, *args, **kwargs):
        if log_method is None:
            print(prefix + body)
        else:
            kwargs["frame"] = LoggingMixin._get_func_prefix(kwargs.pop("frame", None), False)
            log_method(body, prefix, *args, **kwargs)

    @staticmethod
    def merge_logs_by_time(*log_files, tgt_file):
        tgt_folder = os.path.dirname(tgt_file)
        date_str_len = len(datetime.datetime.today().strftime(LoggingMixin._date_format_string_)) + 4
        with lock_manager(tgt_folder, [tgt_file], clear_stuffs_after_exc=False):
            msg_dict, msg_block, last_searched = {}, [], None
            for log_file in log_files:
                with open(log_file, "r") as f:
                    for line in f:
                        date_str = line[:date_str_len]
                        if date_str[:2] == "[ " and date_str[-2:] == " ]":
                            searched_time = datetime.datetime.strptime(
                                date_str[2:-2], LoggingMixin._date_format_string_)
                        else:
                            msg_block.append(line)
                            continue
                        if last_searched is not None:
                            msg_dict.setdefault(last_searched, []).append("".join(msg_block))
                        last_searched = searched_time
                        msg_block = [line]
                    if msg_block:
                        msg_dict.setdefault(last_searched, []).append("".join(msg_block))
            with open(tgt_file, "w") as f:
                f.write("".join(["".join(msg_dict[key]) for key in sorted(msg_dict)]))

    @classmethod
    def start_timer(cls, name):
        cls._time_cache_dict_[name] = time.time()

    @classmethod
    def end_timer(cls, name):
        start_time = cls._time_cache_dict_.pop(name, None)
        if start_time is None:
            raise KeyError(f"'{name}' was not found in time cache dict, "
                           "this may cause by not calling `start_timer` method")
        incrementer = cls._timing_dict_.setdefault(name, Incrementer())
        incrementer.update(time.time() - start_time)

    def log_timing(self):
        timing_str_list = ["=" * 138]
        for name in sorted(self._timing_dict_.keys()):
            incrementer = self._timing_dict_[name]
            timing_str_list.append(
                f"|   {name:<82s}   | "
                f"{fix_float_to_length(incrementer.mean, 10)} ± "
                f"{fix_float_to_length(incrementer.std, 10)} | "
                f"{incrementer.n_record:>12d} hits   |"
            )
            timing_str_list.append("-" * 138)
        self.log_block_msg(
            "\n".join(timing_str_list), title="timing",
            verbose_level=None, msg_level=logging.DEBUG
        )
        return self


class PureLoggingMixin:
    """
    Mixin class to provide (pure) logging method for base class

    Attributes
    ----------
    _loggers_ : dict(int, logging.Logger)
        Recorded all loggers initialized

    _formatter_ : _Formatter
        Formatter for all loggers

    Methods
    ----------
    log_msg(self, name, msg, msg_level=logging.INFO)
        Log something to a file, with logger initialized by `name`

    log_block_msg(self, name, title, body, msg_level=logging.INFO)
        Almost the same as `log_msg`, except adding `title` on top of `body`

    """

    _name = _meta_name = None

    _formatter_ = LoggingMixin._formatter_
    _loggers_: Dict[str, logging.Logger] = {}
    _logger_paths_: Dict[str, str] = {}
    _timing_dict_ = {}

    @property
    def meta_suffix(self):
        return "" if self._meta_name is None else self._meta_name

    @property
    def name_suffix(self):
        return "" if self._name is None else f"-{self._name}"

    @property
    def meta_log_name(self):
        return f"__meta__{self.meta_suffix}{self.name_suffix}"

    @staticmethod
    def get_logging_path(logger):
        logging_path = None
        for handler in logger.handlers:
            if isinstance(handler, logging.FileHandler):
                logging_path = handler.baseFilename
                break
        if logging_path is None:
            raise ValueError(f"No FileHandler was found in given logger '{logger}'")
        return logging_path

    def _get_logger_info(self, name):
        logger = name if isinstance(name, logging.Logger) else self._loggers_.get(name)
        if logger is None:
            raise ValueError(f"logger for '{name}' is not defined, please call `_setup_logger` first")
        if isinstance(name, str):
            logging_path = self._logger_paths_[name]
        else:
            logging_path = self.get_logging_path(logger)
        return logger, os.path.dirname(logging_path), logging_path

    def _setup_logger(self, name, logging_path, level=logging.DEBUG):
        if name in self._loggers_:
            return
        console = logging.StreamHandler()
        console.setLevel(logging.CRITICAL)
        console.setFormatter(self._formatter_)
        file_handler = logging.FileHandler(logging_path)
        file_handler.setFormatter(self._formatter_)
        file_handler.setLevel(level)
        logger = logging.getLogger(name)
        logger.setLevel(logging.DEBUG)
        LoggingMixin._release_handlers(logger)
        logger.addHandler(console)
        logger.addHandler(file_handler)
        PureLoggingMixin._loggers_[name] = logger
        PureLoggingMixin._logger_paths_[name] = logging_path
        for handler in logging.getLogger().handlers:
            handler.setLevel(logging.CRITICAL)
        self.log_block_msg(name, "system version", sys.version)

    def _log_meta_msg(self, msg, msg_level=logging.INFO, frame=None):
        if frame is None:
            frame = inspect.currentframe().f_back
        self.log_msg(self.meta_log_name, msg, msg_level, frame)

    def _log_with_meta(self, task_name, msg, msg_level=logging.INFO, frame=None):
        if frame is None:
            frame = inspect.currentframe().f_back
        self._log_meta_msg(f"{task_name} {msg}", msg_level, frame)
        self.log_msg(task_name, f"current task {msg}", msg_level, frame)

    def log_msg(self, name, msg, msg_level=logging.INFO, frame=None):
        logger, logging_folder, logging_path = self._get_logger_info(name)
        with lock_manager(logging_folder, [logging_path], clear_stuffs_after_exc=False):
            logger.log(msg_level, msg, extra={"func_prefix": LoggingMixin._get_func_prefix(frame)})
        return logger

    def log_block_msg(self, name, title, body, msg_level=logging.INFO, frame=None):
        frame = LoggingMixin._get_func_prefix(frame, False)
        self.log_msg(name, f"{title}\n{body}\n", msg_level, frame)

    def exception(self, name, msg, frame=None):
        logger, logging_folder, logging_path = self._get_logger_info(name)
        with lock_manager(logging_folder, [logging_path], clear_stuffs_after_exc=False):
            logger.exception(msg, extra={"func_prefix": LoggingMixin._get_func_prefix(frame)})

    def del_logger(self, name):
        logger = self.log_msg(name, f"clearing up logger information of '{name}'")
        del self._loggers_[name], self._logger_paths_[name]
        LoggingMixin._release_handlers(logger)
        del logger


class SavingMixin(LoggingMixin):
    """
    Mixin class to provide logging & saving method for base class

    Methods
    ----------
    save(self, folder)
        Save `self` to folder

    def load(self, folder)
        Load from folder

    """

    @property
    @abstractmethod
    def data_tuple_base(self) -> Type[NamedTuple]:
        pass

    @property
    @abstractmethod
    def data_tuple_attributes(self) -> List[str]:
        pass

    @property
    def cache_excludes(self):
        return set()

    @property
    def lock_verbose(self):
        verbose_level = getattr(self, "_verbose_level", None)
        if verbose_level is None:
            return False
        return verbose_level >= 5

    def _data_tuple_context(self, *, is_saving):
        return data_tuple_saving_controller(self, is_saving=is_saving)

    def save(self, folder: str, *, compress: bool = True):
        with self._data_tuple_context(is_saving=True):
            Saving.save_instance(self, folder, self.log_msg)
        if compress:
            abs_folder = os.path.abspath(folder)
            base_folder = os.path.dirname(abs_folder)
            with lock_manager(base_folder, [folder]):
                Saving.compress(abs_folder, remove_original=True)
        return self

    def load(self, folder: str, *, compress: bool = True):
        base_folder = os.path.dirname(os.path.abspath(folder))
        with lock_manager(base_folder, [folder]):
            with Saving.compress_loader(self, folder, compress, remove_extracted=True):
                with self._data_tuple_context(is_saving=False):
                    Saving.load_instance(self, folder, log_method=self.log_msg)
        return self


class Saving(LoggingMixin):
    """
    Util class for saving instances

    Methods
    ----------
    save_instance(instance, folder, log_method=None)
        Save instance to `folder`
        * instance : object, instance to save
        * folder : str, folder to save to
        * log_method : {None, function}, used as `log_method` parameter in
        `log_with_external_method` method of `LoggingMixin`

    load_instance(instance, folder, log_method=None)
        Load instance from `folder`
        * instance : object, instance to load, need to be initialized
        * folder : str, folder to load from
        * log_method : {None, function}, used as `log_method` parameter in
        `log_with_external_method` method of `LoggingMixin`

    """

    delim = "^_^"
    dill_suffix = ".pkl"
    array_sub_folder = "__arrays"

    @staticmethod
    def _check_core(elem):
        if isinstance(elem, dict):
            if not Saving._check_dict(elem):
                return False
        if isinstance(elem, (list, tuple)):
            if not Saving._check_list_and_tuple(elem):
                return False
        if not Saving._check_elem(elem):
            return False
        return True

    @staticmethod
    def _check_elem(elem):
        if isinstance(elem, (type, np.generic, np.ndarray)):
            return False
        if callable(elem):
            return False
        return True

    @staticmethod
    def _check_list_and_tuple(arr: Union[list, tuple]):
        for elem in arr:
            if not Saving._check_core(elem):
                return False
        return True

    @staticmethod
    def _check_dict(d: dict):
        for v in d.values():
            if not Saving._check_core(v):
                return False
        return True

    @staticmethod
    def save_dict(d: dict, name: str, folder: str):
        if Saving._check_dict(d):
            kwargs = {}
            suffix, method, mode = ".json", json.dump, "w"
        else:
            kwargs = {"recurse": True}
            suffix, method, mode = Saving.dill_suffix, dill.dump, "wb"
        with open(os.path.join(folder, f"{name}{suffix}"), mode) as f:
            method(d, f, **kwargs)

    @staticmethod
    def load_dict(name: str, folder: str = None):
        if folder is None:
            folder, name = os.path.split(name)
        name, suffix = os.path.splitext(name)
        if not suffix:
            json_file = os.path.join(folder, f"{name}.json")
            if os.path.isfile(json_file):
                with open(json_file, "r") as f:
                    return json.load(f)
            dill_file = os.path.join(folder, f"{name}{Saving.dill_suffix}")
            if os.path.isfile(dill_file):
                with open(dill_file, "rb") as f:
                    return dill.load(f)
        else:
            assert_msg = f"suffix should be either 'json' or 'pkl', {suffix} found"
            assert suffix in {".json", ".pkl"}, assert_msg
            name = f"{name}{suffix}"
            file = os.path.join(folder, name)
            if os.path.isfile(file):
                if suffix == ".json":
                    mode, load_method = "r", json.load
                else:
                    mode, load_method = "rb", dill.load
                with open(file, mode) as f:
                    return load_method(f)
        raise ValueError(f"config '{name}' is not found under '{folder}' folder")

    @staticmethod
    def deep_copy_dict(d: dict):
        tmp_folder = os.path.join(os.getcwd(), "___tmp_dict_cache___")
        if os.path.isdir(tmp_folder):
            shutil.rmtree(tmp_folder)
        os.makedirs(tmp_folder)
        dict_name = "deep_copy"
        Saving.save_dict(d, dict_name, tmp_folder)
        loaded_dict = Saving.load_dict(dict_name, tmp_folder)
        shutil.rmtree(tmp_folder)
        return loaded_dict

    @staticmethod
    def get_cache_file(instance):
        return f"{type(instance).__name__}.pkl"

    @staticmethod
    def save_instance(instance, folder, log_method=None):
        instance_str = str(instance)
        Saving.log_with_external_method(
            f"saving '{instance_str}' to '{folder}'", Saving.info_prefix, log_method, 5)
        def _record_array(k, v):
            extension = ".npy" if isinstance(v, np.ndarray) else ".lst"
            array_attribute_dict[f"{k}{extension}"] = v
        def _check_array(attr_key_, attr_value_, depth=0):
            if isinstance(attr_value_, dict):
                for k in list(attr_value_.keys()):
                    v = attr_value_[k]
                    extended_k = f"{attr_key_}{delim}{k}"
                    if isinstance(v, dict):
                        _check_array(extended_k, v, depth + 1)
                    elif isinstance(v, array_types):
                        _record_array(extended_k, v)
                        attr_value_.pop(k)
            if isinstance(attr_value_, array_types):
                _record_array(attr_key_, attr_value_)
                if depth == 0:
                    cache_excludes.add(attr_key_)
        main_file, instance_dict = Saving.get_cache_file(instance), shallow_copy_dict(instance.__dict__)
        verbose, cache_excludes = map(
            getattr, [instance] * 2,
            ["lock_verbose", "cache_excludes"], [False, set()]
        )
        if os.path.isdir(folder):
            if verbose:
                prefix = Saving.warning_prefix
                msg = f"'{folder}' will be cleaned up when saving '{instance_str}'"
                Saving.log_with_external_method(msg, prefix, log_method, msg_level=logging.WARNING)
            shutil.rmtree(folder)
        save_path = os.path.join(folder, main_file)
        array_folder = os.path.join(folder, Saving.array_sub_folder)
        tuple(map(lambda folder_: os.makedirs(folder_, exist_ok=True), [folder, array_folder]))
        sorted_attributes, array_attribute_dict = sorted(instance_dict), {}
        delim, array_types = Saving.delim, (list, np.ndarray)
        for attr_key in sorted_attributes:
            if attr_key in cache_excludes:
                continue
            attr_value = instance_dict[attr_key]
            _check_array(attr_key, attr_value)
        cache_excludes.add("_verbose_level_")
        with lock_manager(folder, [os.path.join(folder, main_file)], name=instance_str):
            with open(save_path, "wb") as f:
                dill.dump({k: v for k, v in instance_dict.items() if k not in cache_excludes}, f, recurse=True)
        if array_attribute_dict:
            sorted_array_files = sorted(array_attribute_dict)
            sorted_array_files_full_path = list(map(lambda f_: os.path.join(array_folder, f_), sorted_array_files))
            with lock_manager(array_folder, sorted_array_files_full_path, name=f"{instance_str} (arrays)"):
                for array_file, array_file_full_path in zip(sorted_array_files, sorted_array_files_full_path):
                    array_value = array_attribute_dict[array_file]
                    if array_file.endswith(".npy"):
                        np.save(array_file_full_path, array_value)
                    elif array_file.endswith(".lst"):
                        with open(array_file_full_path, "wb") as f:
                            np.save(f, array_value)
                    else:
                        raise ValueError(f"unrecognized file type '{array_file}' occurred")

    @staticmethod
    def load_instance(instance, folder, *, log_method=None):
        Saving.log_with_external_method(
            f"loading '{instance}' from '{folder}'", Saving.info_prefix, log_method, 5)
        with open(os.path.join(folder, Saving.get_cache_file(instance)), "rb") as f:
            instance.__dict__.update(dill.load(f))
        delim, array_folder = Saving.delim, os.path.join(folder, Saving.array_sub_folder)
        for array_file in os.listdir(array_folder):
            attr_name, attr_ext = os.path.splitext(array_file)
            if attr_ext == ".npy":
                load_method = np.load
            elif attr_ext == ".lst":
                def load_method(path):
                    return np.load(path).tolist()
            else:
                raise ValueError(f"unrecognized file type '{array_file}' occurred")
            array_value = load_method(os.path.join(array_folder, array_file))
            attr_hierarchy = attr_name.split(delim)
            if len(attr_hierarchy) == 1:
                instance.__dict__[attr_name] = array_value
            else:
                hierarchy_dict = instance.__dict__
                for attr in attr_hierarchy[:-1]:
                    hierarchy_dict = hierarchy_dict.setdefault(attr, {})
                hierarchy_dict[attr_hierarchy[-1]] = array_value

    @staticmethod
    def prepare_folder(instance, folder):
        if os.path.isdir(folder):
            instance.log_msg(
                f"'{folder}' already exists, it will be cleared up to save our model",
                instance.warning_prefix, msg_level=logging.WARNING
            )
            shutil.rmtree(folder)
        os.makedirs(folder)

    @staticmethod
    def compress(abs_folder, remove_original=True):
        shutil.make_archive(abs_folder, "zip", abs_folder)
        if remove_original:
            shutil.rmtree(abs_folder)

    @staticmethod
    def compress_loader(instance, folder, is_compress, *, remove_extracted=True):
        class _manager(context_error_handler):
            def __enter__(self):
                if is_compress:
                    if os.path.isdir(folder):
                        instance.log_msg(
                            f"'{folder}' already exists, it will be cleared up to load our model",
                            instance.warning_prefix, msg_level=logging.WARNING
                        )
                        shutil.rmtree(folder)
                    with zipfile.ZipFile(f"{folder}.zip", "r") as zip_ref:
                        zip_ref.extractall(folder)

            def _normal_exit(self, exc_type, exc_val, exc_tb):
                if is_compress and remove_extracted:
                    shutil.rmtree(folder)
        return _manager()


# contexts

class context_error_handler:
    """ Util class which provides exception handling when using context manager """

    @property
    def exception_suffix(self):
        return ""

    def _normal_exit(self, exc_type, exc_val, exc_tb):
        pass

    def _exception_exit(self, exc_type, exc_val, exc_tb):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not exc_type:
            self._normal_exit(exc_type, exc_val, exc_tb)
        else:
            self._exception_exit(exc_type, exc_val, exc_tb)


class timeit(context_error_handler):
    """
    Timing context manager

    Examples
    --------
    >>> with timeit("something"):
    >>>     # do something here
    >>> # will print "~~~  [ info ] timing for    something     : x.xxxx"

    """

    def __init__(self, msg, precision=6):
        self._msg = msg
        self._p = precision

    def __enter__(self):
        self._t = time.time()

    def _normal_exit(self, exc_type, exc_val, exc_tb):
        prefix = LoggingMixin.info_prefix
        print(f"{prefix}timing for {self._msg:^16s} : {time.time() - self._t:{self._p}.{self._p-2}f}")


class _lock_file_refresher(threading.Thread):
    def __init__(self, lock_file, delay=1, refresh=0.01):
        super().__init__()
        self.__stop_event = threading.Event()
        self._lock_file, self._delay, self._refresh = lock_file, delay, refresh
        with open(lock_file, "r") as f:
            self._lock_file_contents = f.read()

    def run(self) -> None:
        counter = 0
        while True:
            counter += 1
            time.sleep(self._refresh)
            if counter * self._refresh >= self._delay:
                counter = 0
                with open(self._lock_file, "w") as f:
                    prefix = "\n\n"
                    add_line = f"{prefix}refreshed at {timestamp()}"
                    f.write(self._lock_file_contents + add_line)
            if self.__stop_event.is_set():
                break

    def stop(self):
        self.__stop_event.set()


class lock_manager(context_error_handler, LoggingMixin):
    """
    Util class to make simultaneously-write process safe with some hacked (ugly) tricks

    Examples
    --------
    >>> import dill
    >>> workplace = "_cache"
    >>> target_write_files_full_path = [
    >>>     os.path.join(workplace, "file1.pkl"),
    >>>     os.path.join(workplace, "file2.pkl")
    >>> ]
    >>> with lock_manager(workplace, target_write_files_full_path):
    >>>     for write_file_full_path in target_write_files_full_path:
    >>>         with open(write_file_full_path, "wb") as wf:
    >>>             dill.dump(..., wf)

    """

    delay = 0.01
    __lock__ = "__lock__"

    def __init__(self, workplace, stuffs, verbose_level=None,
                 set_lock=True, clear_stuffs_after_exc=True, name=None, wait=1000):
        self._workplace = workplace
        self._verbose_level = verbose_level
        self._name, self._wait = name, wait
        os.makedirs(workplace, exist_ok=True)
        self._stuffs, self._set_lock = stuffs, set_lock
        self._clear_stuffs = clear_stuffs_after_exc
        self._is_locked = False

    def __enter__(self):
        frame = inspect.currentframe().f_back
        self.log_msg(
            f"waiting for lock at {self.lock_file}",
            self.info_prefix, 5, logging.DEBUG, frame
        )
        enter_time = file_modify = None
        while True:
            try:
                fd = os.open(self.lock_file, os.O_CREAT | os.O_EXCL | os.O_RDWR)
                self.log_msg("lock acquired", self.info_prefix, 5, logging.DEBUG, frame)
                if not self._set_lock:
                    self.log_msg(
                        "releasing lock since set_lock=False",
                        self.info_prefix, 5, logging.DEBUG, frame
                    )
                    os.unlink(self.lock_file)
                    self.__refresher = None
                else:
                    self.log_msg("writing info to lock file", self.info_prefix, 5, logging.DEBUG, frame)
                    with os.fdopen(fd, "a") as f:
                        f.write(f"name      : {self._name}\n"
                                f"timestamp : {timestamp()}\n"
                                f"workplace : {self._workplace}\n"
                                f"stuffs    :\n{self.cache_stuffs_str}")
                    self.__refresher = _lock_file_refresher(self.lock_file)
                    self.__refresher.start()
                break
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise
                try:
                    if file_modify is None:
                        enter_time = time.time()
                        file_modify = os.path.getmtime(self.lock_file)
                    else:
                        new_file_modify = os.path.getmtime(self.lock_file)
                        if new_file_modify != file_modify:
                            enter_time = time.time()
                            file_modify = new_file_modify
                        else:
                            wait_time = time.time() - enter_time
                            if wait_time >= self._wait:
                                raise ValueError(
                                    f"'{self.lock_file}' has been waited for too long ({wait_time})")
                    time.sleep(random.random() * self.delay + self.delay)
                except ValueError:
                    self.exception(f"lock_manager was blocked by dead lock ({self.lock_file})")
                    raise
                except FileNotFoundError:
                    pass
        self.log_block_msg(
            self.cache_stuffs_str, title="start processing following stuffs:",
            verbose_level=5, msg_level=logging.DEBUG, frame=frame
        )
        self._is_locked = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.__refresher is not None:
            self.__refresher.stop()
            self.__refresher.join()
        if self._set_lock:
            super().__exit__(exc_type, exc_val, exc_tb)

    def _normal_exit(self, exc_type, exc_val, exc_tb, frame=None):
        if self._set_lock:
            os.unlink(self.lock_file)
        if frame is None:
            frame = inspect.currentframe().f_back.f_back.f_back
        self.log_msg("lock released", self.info_prefix, 5, logging.DEBUG, frame)

    def _exception_exit(self, exc_type, exc_val, exc_tb):
        frame = inspect.currentframe().f_back.f_back.f_back
        if self._clear_stuffs:
            for stuff in self._stuffs:
                if os.path.isfile(stuff):
                    self.log_msg(f"clearing cached file: {stuff}", "~~~~~~ ", 5, logging.ERROR, frame)
                    os.remove(stuff)
                elif os.path.isdir(stuff):
                    self.log_msg(f"clearing cached directory: {stuff}", "~~~~~~ ", 5, logging.ERROR, frame)
                    shutil.rmtree(stuff)
        self._normal_exit(exc_type, exc_val, exc_tb, frame)

    @property
    def locked(self):
        return self._is_locked

    @property
    def available(self):
        return not os.path.isfile(self.lock_file)

    @property
    def cache_stuffs_str(self):
        return "\n".join([f"~~~~~~ {stuff}" for stuff in self._stuffs])

    @property
    def exception_suffix(self):
        return f", clearing caches for safety{self.logging_suffix}"

    @property
    def lock_file(self):
        return os.path.join(self._workplace, self.__lock__)

    @property
    def logging_suffix(self):
        return "" if self._name is None else f" - {self._name}"


class batch_manager(context_error_handler):
    """
    Process data in batch

    Parameters
    ----------
    inputs : tuple(np.ndarray), auxiliary array inputs.
    n_elem : {int, float}, indicates how many elements will be processed in a batch
    batch_size : int, indicates the batch_size; if None, batch_size will be calculated by n_elem

    Examples
    --------
    >>> instance = type("test", (object,), {})()
    >>> with batch_manager(instance, np.arange(5), np.arange(1, 6), batch_size=2) as manager:
    >>>     for arr, tensor in manager:
    >>>         print(arr, tensor)
    >>>         # Will print:
    >>>         #   [0 1], [1 2]
    >>>         #   [2 3], [3 4]
    >>>         #   [4]  , [5]

    """

    def __init__(self, *inputs, n_elem=1e6, batch_size=None, max_batch_size=1024):
        if not inputs:
            raise ValueError("inputs should be provided in general_batch_manager")
        input_lengths = list(map(len, inputs))
        self._n, self._rs, self._inputs = input_lengths[0], [], inputs
        assert all(length == self._n for length in input_lengths), "inputs should be of same length"
        if batch_size is not None:
            self._batch_size = batch_size
        else:
            n_elem = int(n_elem)
            self._batch_size = int(n_elem / sum(map(lambda arr: prod(arr.shape[1:]), inputs)))
        self._batch_size = min(max_batch_size, min(self._n, self._batch_size))
        self._n_epoch = int(self._n / self._batch_size)
        self._n_epoch += int(self._n_epoch * self._batch_size < self._n)

    def __enter__(self):
        return self

    def __iter__(self):
        self._start, self._end = 0, self._batch_size
        return self

    def __next__(self):
        if self._start >= self._n:
            raise StopIteration
        batched_data = tuple(map(lambda arr: arr[self._start:self._end], self._inputs))
        self._start, self._end = self._end, self._end + self._batch_size
        if len(batched_data) == 1:
            return batched_data[0]
        return batched_data

    def __len__(self):
        return self._n_epoch


class timing_context(context_error_handler):
    """
    Wrap codes in any base class of `LoggingMixin` with this timing context to timeit

    Parameters
    ----------
    logging_mixin : LoggingMixin, arbitrary base classes of LoggingMixin
    name : str, explain what the wrapped codes are doing

    Examples
    --------
    >>> import time
    >>> import random
    >>> instance = type(
    >>>    "test", (LoggingMixin,),
    >>>    {"config": {}, "_verbose_level": 2}
    >>> )()._init_logging(2, True)
    >>> for _ in range(50):
    >>>     with timing_context(instance, "random sleep"):
    >>>         time.sleep(random.random() * 0.1)
    >>> instance.log_timing()

    """

    def __init__(self, logging_mixin, name):
        self._cls, self._name = logging_mixin, name

    @property
    def timer_name(self):
        return f"[{type(self._cls).__name__:^24s}] {self._name}"

    def __enter__(self):
        self._cls.start_timer(self.timer_name)

    def _normal_exit(self, exc_type, exc_val, exc_tb):
        self._cls.end_timer(self.timer_name)


class data_tuple_saving_controller(context_error_handler):
    """
    Help saving DataTuple of SavingMixin

    Parameters
    ----------
    instance : SavingMixin, instance whose DataTuples are to be saved / loaded
    is_saving : bool, whether it is a saving context or not
    """

    __prefix__ = "_Data_Tuple__"

    def __init__(self,
                 instance: SavingMixin,
                 *, is_saving: bool):
        self._instance = instance
        self._is_saving = is_saving
        self._data_tuple_base = instance.data_tuple_base
        self._data_tuple_attributes = instance.data_tuple_attributes
        if self.trigger and is_saving:
            self._data_tuples: List[NamedTuple] = [
                instance.__dict__.pop(attr) for attr in self._data_tuple_attributes]

    def __enter__(self):
        if self.trigger and self._is_saving:
            self.__tmp_attr_list = []
            for attr, data_tuple in zip(self._data_tuple_attributes, self._data_tuples):
                local_attr_list = self._get_attr(attr, data_tuple)
                for local_attr, data in zip(local_attr_list, data_tuple):
                    setattr(self._instance, local_attr, data)
                self.__tmp_attr_list += local_attr_list

    @property
    def trigger(self):
        return self._data_tuple_base is not None and self._data_tuple_attributes is not None

    def _normal_exit(self, exc_type, exc_val, exc_tb):
        if self.trigger:
            if self._is_saving:
                for attr, data_tuple in zip(self._data_tuple_attributes, self._data_tuples):
                    setattr(self._instance, attr, self._data_tuple_base(*data_tuple))
                for attr in self.__tmp_attr_list:
                    self._instance.__dict__.pop(attr)
                del self._data_tuples
            else:
                attr_pool_map = self._get_attr()
                for core_attr, attr_dict in attr_pool_map.items():
                    local_attr_values = []
                    for idx in range(len(attr_dict)):
                        local_attr = attr_dict[idx]
                        local_attr_values.append(self._instance.__dict__.pop(local_attr))
                    setattr(self._instance, core_attr, self._data_tuple_base(*local_attr_values))

    def _get_attr(self,
                  attr: str = None,
                  data_tuple: NamedTuple = None) -> Union[None, List[str], Dict[str, Dict[int, str]]]:
        prefix = self.__prefix__
        if self._is_saving:
            if data_tuple is None:
                raise ValueError("data tuple should be provided in saving context")
            assert isinstance(attr, str), "attr should be string in saving context"
            return [f"{prefix}{attr}_{i}" for i in range(len(data_tuple))]
        attr_pool = list(filter(lambda attr_: attr_.startswith(prefix), self._instance.__dict__.keys()))
        attr_pool_split = [attr_[len(prefix):].split("_") for attr_ in attr_pool]
        attr_pool_map = {}
        for attr, attr_split in zip(attr_pool, attr_pool_split):
            core_attr, idx = "_".join(attr_split[:-1]), int(attr_split[-1])
            local_map = attr_pool_map.setdefault(core_attr, {})
            local_map[idx] = attr
        loaded_attr_list = sorted(attr_pool_map)
        preset_attr_list = sorted(self._data_tuple_attributes)
        assert_msg = f"loaded attributes ({loaded_attr_list}) " \
                     f"are not identical with preset attributes ({preset_attr_list})"
        assert loaded_attr_list == preset_attr_list, assert_msg
        return attr_pool_map


__all__ = [
    "get_indices_from_another", "get_unique_indices", "get_one_hot", "hash_code", "prefix_dict",
    "timestamp", "fix_float_to_length", "truncate_string_to_length", "grouped", "is_numeric", "update_dict",
    "show_or_save", "show_or_return", "Saving", "LoggingMixin", "PureLoggingMixin", "SavingMixin",
    "context_error_handler", "timeit", "lock_manager", "batch_manager", "timing_context", "prod",
    "get_counter_from_arr", "shallow_copy_dict", "register_core", "data_tuple_saving_controller"
]



import os
import dill
import time
import pprint
import random
import signal
import inspect
import logging
import platform

from typing import *
from pathos.pools import ProcessPool
from multiprocessing import Process
from multiprocessing.managers import SyncManager

from ..misc import *
from ..manage import *

WINDOWS = platform.system() == "Windows"
dill._dill._reverse_typemap["ClassType"] = type


class Parallel(PureLoggingMixin):
    """
    Util class which can help running tasks in parallel

    Warnings
    ----------
    On Windows platform, functions are dramatically reduced because Windows does not well support pickling.
    * In this occasion, `Parallel` will simply leverage `pathos` to do the jobs

    Parameters
    ----------
    num_jobs : int, number of jobs run in parallel
    sleep : float, idle duration of new jobs
    use_cuda: bool, whether tasks need CUDA or not
    name : str, summary name of these tasks
    meta_name : str, name of the meta information
    logging_folder : str, where the logging will be placed
    task_names : List[str], names of each task
    resource_config : Dict[str, Any], config used in `ResourceManager`

    Examples
    ----------
    >>> def add_one(x):
    >>>     import time
    >>>     time.sleep(1)
    >>>     return x + 1
    >>>
    >>> print(Parallel(10)(add_one, list(range(10)))._rs)

    """

    class _ParallelError(Exception):
        pass

    def __init__(self,
                 num_jobs: int,
                 *,
                 sleep: float = 1.,
                 use_cuda: bool = False,
                 name: str = None,
                 meta_name: str = None,
                 logging_folder: str = None,
                 task_names: List[str] = None,
                 resource_config: Dict[str, Any] = None):
        self._rs = None
        self._n_jobs, self._sleep, self._use_cuda = num_jobs, sleep, use_cuda
        if resource_config is None:
            resource_config = {}
        if logging_folder is None:
            logging_folder = os.path.join(os.getcwd(), "_parallel_", "logs")
        self._resource_config = resource_config
        self._name, self._meta_name = name, meta_name
        self._logging_folder, self._task_names = logging_folder, task_names
        self._refresh_patience = resource_config.setdefault("refresh_patience", 10)
        self._init_logger(self.meta_log_name)

    def __call__(self, f, *args_list) -> "Parallel":
        #   if f returns a dict with 'terminate' key, Parallel can be terminated at early stage by
        # setting 'terminate' key to True
        n_tasks = len(args_list[0])
        n_jobs = min(self._n_jobs, n_tasks)
        if WINDOWS:
            p = ProcessPool(ncpus=n_jobs)
            task_names = list(map(self._get_task_name, range(n_tasks)))
            results = p.map(f, *args_list)
            self._rs = dict(zip(task_names, results))
            return self
        self._func, self._args_list = f, args_list
        self._cursor, self._all_task_ids = 0, list(range(n_jobs, n_tasks))
        self._log_meta_msg("initializing sync manager")
        self._sync_manager = SyncManager()
        self._sync_manager.start(lambda: signal.signal(signal.SIGINT, signal.SIG_IGN))
        self._rs = self._sync_manager.dict({
            "__meta__": {"n_jobs": n_jobs, "n_tasks": n_tasks, "terminated": False},
            "__exceptions__": {}
        })
        self._overwritten_task_info = {}
        self._pid2task_id = None
        self._log_meta_msg("initializing resource manager")
        self._resource_manager = ResourceManager(
            self._resource_config, self._get_task_name, self._refresh_patience)
        self._log_meta_msg("registering PC manager")
        pc_manager = PCManager()
        ram_methods = {
            "get_pid_usage_dict": None,
            "get_pid_usage": pc_manager.get_pid_ram_usage,
            "get_available_dict": lambda: {"total": pc_manager.get_available_ram()}
        }
        self._resource_manager.register("RAM", ram_methods)
        gpu_config = self._resource_config.setdefault("gpu_config", {})
        default_cuda_list = None if self._use_cuda else []
        available_cuda_list = gpu_config.setdefault("available_cuda_list", default_cuda_list)
        if available_cuda_list is None or available_cuda_list:
            self._log_meta_msg("registering GPU manager")
            if available_cuda_list is not None:
                available_cuda_list = list(map(int, available_cuda_list))
            gpu_manager = GPUManager(available_cuda_list)
            gpu_methods = {
                "get_pid_usage": None,
                "get_pid_usage_dict": gpu_manager.get_pid_usages,
                "get_available_dict": gpu_manager.get_gpu_frees
            }
            self._resource_manager.register("GPU", gpu_methods)
        self._resource_manager.register_logging(self._init_logger, self)
        self._log_meta_msg("initializing with refreshing")
        self._refresh(skip_check_finished=True)
        self._working_processes = None
        try:
            self._log_meta_msg("initializing processes")
            init_task_ids = list(range(n_jobs))
            init_processes = [self._get_process(i, start=False) for i in init_task_ids]
            if self.terminated:
                self._user_terminate()
            init_failed_slots, init_failed_task_ids = [], []
            for i, (task_id, process) in enumerate(zip(init_task_ids, init_processes)):
                if process is None:
                    init_failed_slots.append(i)
                    init_failed_task_ids.append(task_id)
                    task_name = self._get_task_name(task_id)
                    self._log_with_meta(
                        task_name, "initialization failed, it may due to lack of resources",
                        msg_level=logging.WARNING
                    )
            if init_failed_slots:
                for slot in init_failed_slots:
                    init_task_ids[slot] = None
                    init_processes[slot] = [None] * 4
                self._all_task_ids = init_failed_task_ids + self._all_task_ids
            self._working_task_ids = init_task_ids
            self._working_processes, task_info = map(list, zip(*init_processes))
            self._log_meta_msg("starting all initial processes")
            tuple(map(lambda p: None if p is None else p.start(), self._working_processes))
            tuple(map(self._record_process, self._working_processes, self._working_task_ids, task_info))
            self._resource_manager.initialize_running_usages()
            self._log_meta_msg("entering parallel main loop")
            while True:
                self._log_meta_msg("waiting for finished slot")
                self._wait_and_handle_finish(wait_until_finish=True)
                if not self._add_new_processes():
                    break
        except KeyboardInterrupt:
            self.exception(self.meta_log_name, f"keyboard interrupted")
            exceptions = self.exceptions
            exceptions["base"] = self._ParallelError("Keyboard Interrupted")
            self._rs["__exceptions__"] = exceptions
        except Exception as err:
            self.exception(self.meta_log_name, f"exception occurred, {err}")
            exceptions = self.exceptions
            exceptions["base"] = err
            self._rs["__exceptions__"] = exceptions
        finally:
            self._log_meta_msg("joining processes left behind")
            if self._working_processes is not None:
                for process in self._working_processes:
                    if process is None:
                        continue
                    process.join()
            self._log_meta_msg("casting parallel results to Python dict")
            self._rs = dict(self._rs)
            self._log_meta_msg("shutting down sync manager")
            self._sync_manager.shutdown()
            self.log_block_msg(
                self.meta_log_name, "parallel results",
                pprint.pformat(self._rs, compact=True)
            )
        return self

    @property
    def meta(self):
        return self._rs["__meta__"]

    @property
    def exceptions(self):
        return self._rs["__exceptions__"]

    @property
    def terminated(self):
        return self.meta["terminated"]

    @property
    def parallel_results(self):
        return self._rs

    def __sleep(self, skip_check_finished):
        time.sleep(self._sleep + random.random())
        self._refresh(skip_check_finished=skip_check_finished)

    def __wait(self, wait_until_finished):
        # should return a sorted list
        try:
            while True:
                self._log_meta_msg(
                    "waiting for slots (working tasks : "
                    f"{', '.join(map(self._get_task_name, filter(bool, self._working_task_ids)))})",
                    msg_level=logging.DEBUG
                )
                finished_slots = []
                for i, (task_id, process) in enumerate(zip(
                        self._working_task_ids, self._working_processes)):
                    if process is None:
                        self._log_meta_msg(f"pending on slot {i}")
                        finished_slots.append(i)
                        continue
                    task_name = self._get_task_name(task_id)
                    if not process.is_alive():
                        self._log_with_meta(task_name, f"in slot {i} is found finished")
                        finished_slots.append(i)
                if not wait_until_finished or finished_slots:
                    return finished_slots
                self.__sleep(skip_check_finished=True)
        except KeyboardInterrupt:
            self._set_terminate(scope="wait")
            raise self._ParallelError("Keyboard Interrupted")

    def _init_logger(self, task_name):
        logging_folder = os.path.join(self._logging_folder, task_name)
        os.makedirs(logging_folder, exist_ok=True)
        logging_path = os.path.join(logging_folder, f"{timestamp()}.log")
        self._setup_logger(task_name, logging_path)

    def _refresh(self, skip_check_finished):
        if self._pid2task_id is None:
            self._pid2task_id = self._resource_manager.pid2task_id
        if not self._resource_manager.inference_usages_initialized:
            self._resource_manager.initialize_inference_usages()
        if not self._resource_manager.checkpoint_initialized:
            return
        self._resource_manager.log_pid_usages_and_inference_frees()
        self._resource_manager.check()
        if not skip_check_finished:
            self._wait_and_handle_finish(wait_until_finish=False)

    def _wait_and_handle_finish(self, wait_until_finish):
        finished_slots = self.__wait(wait_until_finish)
        if not finished_slots:
            return
        if self.terminated:
            self._user_terminate()
        finished_bundle = [[], []]
        for finished_slot in finished_slots[::-1]:
            tuple(map(list.append, finished_bundle, map(
                list.pop, [self._working_task_ids, self._working_processes], [finished_slot] * 2)))
        for task_id, process in zip(*finished_bundle):
            task_name = self._resource_manager.handle_finish(process, task_id)
            if task_name is None:
                continue
            self.del_logger(task_name)

    def _add_new_processes(self):
        n_working = len(self._working_processes)
        n_new_jobs = self._n_jobs - n_working
        n_res = len(self._all_task_ids) - self._cursor
        if n_res > 0:
            n_new_jobs = min(n_new_jobs, n_res)
            for _ in range(n_new_jobs):
                new_task_id = self._all_task_ids[self._cursor]
                self._working_processes.append(self._get_process(new_task_id))
                self._working_task_ids.append(new_task_id)
                self._cursor += 1
            return True
        return n_working > 0

    def _user_terminate(self):
        self._log_meta_msg("`_user_terminate` method hit, joining processes", logging.ERROR)
        for process in self._working_processes:
            if process is None:
                continue
            process.join()
        self._log_meta_msg("processes joined, raising self._ParallelError", logging.ERROR)
        recorded_exceptions = self.exceptions
        if not recorded_exceptions:
            raise self._ParallelError("Parallel terminated by user action")
        else:
            raise self._ParallelError("Parallel terminated by unexpected errors")

    def _set_terminate(self, **kwargs):
        meta = self.meta
        meta["terminated"] = True
        self._rs["__meta__"] = meta
        if not kwargs:
            suffix = ""
        else:
            suffix = f" ({' ; '.join(f'{k}: {v}' for k, v in kwargs.items())})"
        self._log_meta_msg(f"`_set_terminate` method hit{suffix}", logging.ERROR)

    def _get_task_name(self, task_id):
        if task_id is None:
            return
        if self._task_names is None:
            task_name = f"task_{task_id}{self.name_suffix}"
        else:
            task_name = f"{self._task_names[task_id]}{self.name_suffix}"
        self._init_logger(task_name)
        return task_name

    def _f_wrapper(self, task_id, cuda=None):
        task_name = self._get_task_name(task_id)
        logger = self._loggers_[task_name]

        def log_method(msg, msg_level=logging.INFO, frame=None):
            if frame is None:
                frame = inspect.currentframe().f_back
            self.log_msg(logger, msg, msg_level, frame)
            return logger

        def _inner(*args):
            if self.terminated:
                return
            try:
                log_method("task started")
                kwargs = {}
                f_wants_cuda = f_wants_log_method = False
                f_signature = inspect.signature(self._func)
                for name, param in f_signature.parameters.items():
                    if param.kind is inspect.Parameter.VAR_KEYWORD:
                        f_wants_cuda = f_wants_log_method = True
                        break
                    if param.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD:
                        if name == "cuda":
                            f_wants_cuda = True
                            continue
                        if name == "log_method":
                            f_wants_log_method = True
                            continue
                if not f_wants_cuda:
                    if self._use_cuda:
                        log_method("task function doesn't want cuda but cuda is used", logging.WARNING)
                else:
                    log_method("task function wants cuda")
                    kwargs["cuda"] = cuda
                if not f_wants_log_method:
                    log_method("task function doesn't want log_method", logging.WARNING)
                else:
                    log_method("task function wants log_method")
                    kwargs["log_method"] = log_method
                self._rs[task_name] = rs = self._func(*args, **kwargs)
                terminate = isinstance(rs, dict) and rs.get("terminate", False)
                if not terminate:
                    log_method("task finished")
            except KeyboardInterrupt:
                log_method("key board interrupted", logging.ERROR)
                return
            except Exception as err:
                logger.exception(
                    f"exception occurred, {err}",
                    extra={"func_prefix": LoggingMixin._get_func_prefix(None)}
                )
                terminate = True
                exceptions = self.exceptions
                self._rs[task_name] = rs = err
                exceptions[task_name] = rs
                self._rs["__exceptions__"] = exceptions
            if terminate:
                self._set_terminate(scope="f_wrapper", task=task_name)
                log_method("task terminated", logging.ERROR)

        return _inner

    def _get_process(self, task_id, start=True):
        rs = self._resource_manager.get_process(
            task_id, lambda: self.__sleep(skip_check_finished=False), start)
        task_name = rs["__task_name__"]
        if not rs["__create_process__"]:
            return
        if not self._use_cuda or "GPU" not in rs:
            args = (task_id,)
        else:
            args = (task_id, rs["GPU"]["tgt_resource_id"])
        target = self._f_wrapper(*args)
        process = Process(target=target, args=tuple(args[task_id] for args in self._args_list))
        self._log_with_meta(task_name, "process created")
        if start:
            process.start()
            self._log_with_meta(task_name, "process started")
            self._record_process(process, task_id, rs)
            return process
        return process, rs

    def _record_process(self, process, task_id, rs):
        if process is None:
            return
        self._resource_manager.record_process(process, task_id, rs)


__all__ = ["Parallel"]



from .ml import *
from .misc import *
from .manage import *


__all__ = [
    "get_indices_from_another", "get_unique_indices", "get_one_hot", "hash_code", "prefix_dict",
    "timestamp", "fix_float_to_length", "truncate_string_to_length", "grouped", "is_numeric", "update_dict",
    "show_or_save", "show_or_return", "Grid", "Saving", "Anneal", "Metrics", "Estimator", "ScalarEMA", "Visualizer",
    "LoggingMixin", "PureLoggingMixin", "SavingMixin", "context_error_handler", "timeit", "lock_manager",
    "batch_manager", "timing_context", "prod", "shallow_copy_dict", "register_core",
    "data_tuple_saving_controller", "PCManager", "GPUManager", "ResourceManager"
]



from setuptools import setup, find_packages

VERSION = "0.1.3"

DESCRIPTION = "Some commonly used functions and modules"
with open("README.md") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name="carefree-toolkit",
    version=VERSION,
    packages=find_packages(exclude=("tests",)),
    install_requires=[
        "pathos", "joblib",
        "dill", "future", "psutil", "pillow",
        "cython>=0.29.12", "numpy>=1.16.2", "scipy>=1.2.1",
        "scikit-learn>=0.20.3", "matplotlib>=3.0.3",
        "mkdocs", "mkdocs-material", "mkdocs-minify-plugin",
        "Pygments", "pymdown-extensions"
    ],
    author="carefree0910",
    author_email="syameimaru_kurumi@pku.edu.cn",
    url="https://github.com/carefree0910/carefree-toolkit",
    download_url=f"https://github.com/carefree0910/carefree-toolkit/archive/v{VERSION}.tar.gz",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    keywords="python numpy data-science"
)



import os
import math
import logging

import numpy as np
import matplotlib.pyplot as plt

from typing import *
from scipy import interp
from scipy import stats as ss
from sklearn import metrics
from itertools import product

from ..misc import *


class Anneal:
    """
    Util class which can provide annealed numbers with given `method`
    * Formulas could be found in `_initialize` method

    Parameters
    ----------
    method : str, indicates which anneal method to be used
    n_iter : int, indicates how much 'steps' will be taken to reach `ceiling` from `floor`
    floor : float, indicates the start point of the annealed number
    ceiling : float, indicates the end point of the annealed number

    Examples
    --------
    >>> anneal = Anneal("linear", 50, 0.01, 0.99)
    >>> for i in range(100):
    >>>     # for i == 0, 1, ..., 48, 49, it will pop 0.01, 0.03, ..., 0.97, 0.99
    >>>     # for i == 50, 51, ..., 98, 99, it will pop 0.99, 0.99, ..., 0.99, 0.99
    >>>     anneal.pop()

    """

    def __init__(self, method, n_iter, floor, ceiling):
        self._n_iter = max(1, n_iter)
        self._method, self._max_iter = method, n_iter
        self._floor, self._ceiling = floor, ceiling
        self._cache = self._rs = self._cursor = 0
        self._initialize()

    def __str__(self):
        return f"Anneal({self._method})"

    __repr__ = __str__

    def _initialize(self):
        n_anneal = max(1, self._n_iter - 1)
        if self._method == "linear":
            self._cache = (self._ceiling - self._floor) / n_anneal
        elif self._method == "log":
            self._cache = (self._ceiling - self._floor) / math.log(n_anneal)
        elif self._method == "quad":
            self._cache = (self._ceiling - self._floor) / (n_anneal ** 2)
        elif self._method == "sigmoid":
            self._cache = 8 / n_anneal
        self._rs = self._floor - self._cache

    def _update_linear(self):
        self._rs += self._cache

    def _update_log(self):
        self._rs = math.log(self._cursor) * self._cache

    def _update_quad(self):
        self._rs = self._cursor ** 2 * self._cache

    def _update_sigmoid(self):
        self._rs = self._ceiling / (1 + math.exp(4 - self._cursor * self._cache))

    def pop(self):
        self._cursor += 1
        if self._cursor >= self._n_iter:
            return self._ceiling
        getattr(self, f"_update_{self._method}")()
        return self._rs

    def visualize(self):
        rs = [self.pop() for _ in range(self._max_iter)]
        plt.figure()
        plt.plot(range(len(rs)), rs)
        plt.show()
        self._initialize()
        return self


class Metrics(LoggingMixin):
    """
    Util class to calculate a whole variety of metrics

    Warnings
    ----------
    * Notice that 2-dimensional arrays are desired, not flattened arrays
    * Notice that first two args of each metric method must be `y` & `pred`

    Parameters
    ----------
    metric_type : str, indicates which kind of metric is to be calculated
    config : dict
        Configuration for the specific metric
        * e.g. for quantile metric, you need to specify which quantile is to be evaluated
    verbose_level : int, verbose level of Metrics

    Examples
    --------
    >>> import numpy as np
    >>> predictions, y_true = map(np.atleast_2d, [[1., 2., 3.], [0., 2., 1.]])
    >>> print(Metrics("mae", {}).score(y_true.T, predictions.T))  # will be 1.

    """

    sign_dict = {
        "ddr": -1, "quantile": -1, "cdf_loss": -1, "loss": -1,
        "f1_score": 1, "r2_score": 1, "auc": 1, "multi_auc": 1,
        "acc": 1, "mae": -1, "mse": -1, "ber": -1,
        "correlation": 1, "top_k_score": 1
    }
    requires_prob_metrics = {"auc", "multi_auc"}
    optimized_binary_metrics = {"acc", "ber"}
    custom_metrics = {}

    def __init__(self, metric_type=None, config=None, verbose_level=None):
        if config is None:
            config = {}
        self.type, self.config, self._verbose_level = metric_type, config, verbose_level

    def __str__(self):
        return f"Metrics({self.type})"

    __repr__ = __str__

    @property
    def sign(self):
        return Metrics.sign_dict[self.type]

    @property
    def use_loss(self):
        return self.type == "loss"

    @property
    def requires_prob(self):
        return self.type in self.requires_prob_metrics

    def _handle_nan(self, y, pred):
        pred_valid_mask = np.all(~np.isnan(pred), axis=1)
        valid_ratio = pred_valid_mask.mean()
        if valid_ratio == 0:
            self.log_msg("all pred are nan", self.error_prefix, 2, logging.ERROR)
            return None, None
        if valid_ratio != 1:
            self.log_msg(
                f"pred contains nan (ratio={valid_ratio:6.4f})",
                self.error_prefix, 2, logging.ERROR
            )
            y, pred = y[pred_valid_mask], pred[pred_valid_mask]
        return y, pred

    @classmethod
    def add_metric(cls, f, name, sign, requires_prob):
        if name in cls.sign_dict:
            print(f"{LoggingMixin.warning_prefix}'{name}' is already registered in Metrics")
        cls.sign_dict[name] = sign
        cls.custom_metrics[name] = {
            "f": f,
            "sign": sign,
            "requires_prob": requires_prob
        }
        if requires_prob:
            cls.requires_prob_metrics.add(name)

    def score(self, y, pred):
        if self.type is None:
            raise ValueError("`score` method was called but type is not specified in Metrics")
        y, pred = self._handle_nan(y, pred)
        if y is None or pred is None:
            return float("nan")
        custom_metric_info = self.custom_metrics.get(self.type)
        if custom_metric_info is not None:
            return custom_metric_info["f"](self, y, pred)
        return getattr(self, self.type)(y, pred)

    # config-dependent metrics

    def quantile(self, y, pred):
        q, error = self.config["q"], y - pred
        if isinstance(q, list):
            q = np.array(q, np.float32).reshape([-1, 1])
        return np.maximum(q * error, (q - 1) * error).mean(0).sum()

    def cdf_loss(self, y, pred, yq=None):
        if yq is None:
            eps = self.config.setdefault("eps", 1e-6)
            mask = y <= self.config["anchor"]
            pred = np.clip(pred, eps, 1 - eps)
            cdf_raw = pred / (1 - pred)
            return -np.mean(mask * cdf_raw - np.log(1 - pred))
        q, self.config["q"] = self.config.get("q"), pred
        loss = self.quantile(y, yq)
        if q is None:
            self.config.pop("q")
        else:
            self.config["q"] = q
        return loss

    # static metrics

    @staticmethod
    def f1_score(y, pred):
        return metrics.f1_score(y.ravel(), pred.ravel())

    @staticmethod
    def r2_score(y, pred):
        return metrics.r2_score(y.ravel(), pred.ravel())

    @staticmethod
    def auc(y, pred):
        n_classes = pred.shape[1]
        if n_classes == 2:
            return metrics.roc_auc_score(y.ravel(), pred[..., 1])
        return Metrics.multi_auc(y, pred)

    @staticmethod
    def multi_auc(y, pred):
        n_classes = pred.shape[1]
        y = get_one_hot(y.ravel(), n_classes)
        fpr, tpr = [None] * n_classes, [None] * n_classes
        for i in range(n_classes):
            fpr[i], tpr[i], _ = metrics.roc_curve(y[..., i], pred[..., i])
        new_fpr = np.unique(np.concatenate([fpr[i] for i in range(n_classes)]))
        new_tpr = np.zeros_like(new_fpr)
        for i in range(n_classes):
            new_tpr += interp(new_fpr, fpr[i], tpr[i])
        new_tpr /= n_classes
        return metrics.auc(new_fpr, new_tpr)

    @staticmethod
    def acc(y, pred):
        return np.mean(y == pred)

    @staticmethod
    def mae(y, pred):
        return np.mean(np.abs(y - pred))

    @staticmethod
    def mse(y, pred):
        return np.mean(np.square(y - pred))

    @staticmethod
    def ber(y, pred):
        mat = metrics.confusion_matrix(y.ravel(), pred.ravel())
        tp = np.diag(mat)
        fp = mat.sum(axis=0) - tp
        fn = mat.sum(axis=1) - tp
        tn = mat.sum() - (tp + fp + fn)
        return 0.5 * np.mean((fn / (tp + fn) + fp / (tn + fp)))

    @staticmethod
    def correlation(y, pred):
        return float(ss.pearsonr(y.ravel(), pred.ravel())[0])

    # auxiliaries

    @staticmethod
    def get_binary_threshold(y, probabilities, metric_type):
        pos_probabilities = probabilities[..., 1]
        fpr, tpr, thresholds = metrics.roc_curve(y, pos_probabilities)
        _, counts = np.unique(y, return_counts=True)
        pos = counts[1] / len(y)
        if metric_type == "ber":
            metric = 0.5 * (1 - tpr + fpr)
        elif metric_type == "acc":
            metric = tpr * pos + (1 - fpr) * (1 - pos)
        else:
            raise NotImplementedError(f"transformation from fpr, tpr -> '{metric_type}' is not implemented")
        metric *= Metrics.sign_dict[metric_type]
        return thresholds[np.argmax(metric)]


class Estimator:
    """
    Util class to estimate the performances of a group of methods, on specific dataset & metric

    Parameters
    ----------
    metric_type : str, indicates which kind of metric is to be calculated
    **kwargs : used to initialize `Metrics` instance

    Examples
    --------
    >>> import numpy as np
    >>> x, y = map(np.atleast_2d, [[1., 2., 3.], [0., 2., 1.]])
    >>> identical = lambda x_: x_
    >>> minus_one = lambda x_: x_ - 1
    >>> # ==========================================================
    >>> # |             identical  |    mae     |  1.000000  |
    >>> # |             minus_one  |    mae     |  0.666667  |  <-
    >>> # ----------------------------------------------------------
    >>> Estimator("mae").estimate(x, y, {"identical": identical, "minus_one": minus_one})

    """

    def __init__(self, metric_type: str, **kwargs):
        self.scores = {}
        self.best_method = None
        self._metric = Metrics(metric_type, **kwargs)

    def __str__(self):
        return f"Estimator({self.type})"

    __repr__ = __str__

    @property
    def type(self) -> str:
        return self._metric.type

    def estimate(self,
                 x: np.ndarray,
                 y: np.ndarray,
                 methods: Dict[str, Union[None, Callable]],
                 *,
                 verbose: bool = True) -> None:
        self.scores = {
            name: None if method is None else self._metric.score(y, method(x))
            for name, method in methods.items()
        }
        msg_list = []
        best_idx, best_score = -1, -math.inf
        sorted_method_names = sorted(self.scores)
        for i, name in enumerate(sorted_method_names):
            score = self.scores[name]
            if score is None:
                continue
            if verbose:
                msg_list.append(f"|  {name:>20s}  |  {self._metric.type:^8s}  |  {score:8.6f}  |")
            new_score = score * self._metric.sign
            if new_score > best_score:
                best_idx, best_score = i, new_score
        self.best_method = sorted_method_names[best_idx]
        if verbose:
            msg_list[best_idx] += "  <-  "
            width = max(map(len, msg_list))
            msg_list.insert(0, "=" * width)
            msg_list.append("-" * width)
            print("\n".join(msg_list))


class ModelPattern(LoggingMixin):
    """
    Util class to create an interface for users to leverage `Comparer` (and more in the future)

    Parameters
    ----------
    init_method : Callable[[], object]
    * If None, then `ModelPattern` will not perform model creation
    * If Callable, then `ModelPattern` will initialize a model with it
    train_method : Callable[[object], None]
    * If None, then `ModelPattern` will not perform model training
    * If Callable, then `ModelPattern` will train the created model (from `init_method`) with it
    predict_method : Union[str, Callable[[np.ndarray], np.ndarray]]
    * If str, then `ModelPattern` will use `getattr` to get the predict method of the model obtained from above
        In this case, `init_method` must be provided (`train_method` is still optional, because you can create a
      trained model in `init_method`)
    * If Callable, then `ModelPattern` will use it for prediction

    Examples
    --------
    >>> import numpy as np
    >>> x, y = map(np.atleast_2d, [[1., 2., 3.], [0., 2., 1.]])
    >>> predict_method = lambda x_: x_ - 1
    >>> init_method = lambda: type("Test", (), {"predict": lambda self, x_: predict_method(x_)})()
    >>> # Will both be [[0., 1., 2.]]
    >>> ModelPattern(init_method=init_method).predict(x)
    >>> ModelPattern(predict_method=predict_method).predict(x)

    """

    def __init__(self,
                 *,
                 init_method: Callable[[], object] = None,
                 train_method: Callable[[object], None] = None,
                 predict_method: Union[str, Callable[[np.ndarray], np.ndarray]] = "predict",
                 trigger_logging: bool = False,
                 verbose_level: int = 2):
        if init_method is None:
            self.model = None
        else:
            self.model = init_method()
        if train_method is not None:
            train_method(self.model)
        self._predict_method = predict_method
        self._init_logging(verbose_level, trigger_logging)

    def predict(self,
                x: np.ndarray) -> np.ndarray:
        predict_method = self._predict_method
        if isinstance(predict_method, str):
            if self.model is None:
                raise ValueError("Either init_method or Callable predict_method is required in ModelPattern")
            predict_method = getattr(self.model, predict_method)
        elif self.model is not None:
            self.log_msg(
                "predict_method is Callable but model is also created, which has no effect",
                self.warning_prefix, 2, logging.WARNING
            )
        return predict_method(x)


class Comparer:
    """
    Util class to compare a group of `ModelPattern`s on a group of estimators

    Parameters
    ----------
    model_patterns : Dict[str, Union[ModelPattern, Dict[str, ModelPattern]]]
    * If values are `ModelPattern`, then all estimators will use this only `ModelPattern` make predictions
    * If values are Dict[str, ModelPattern], then each estimator will use values.get(estimator.type) to
      make predictions. If corresponding `ModelPattern` does not exist (values.get(estimator.type) is None),
      then corresponding estimation will be skipped
    estimators : List[Estimator], list of estimators which we are interested in

    Examples
    --------
    >>> import numpy as np
    >>> x, y = map(np.atleast_2d, [[1., 2., 3.], [0., 2., 1.]])
    >>> identical = lambda x_: x_
    >>> minus_one = lambda x_: x_ - 1
    >>> patterns = {
    >>>     "identical": ModelPattern(predict_method=identical),
    >>>     "minus_one": ModelPattern(predict_method=minus_one)
    >>> }
    >>> estimators = [Estimator("mse"), Estimator("mae")]
    >>> #==========================================================
    >>> # |             identical  |    mse     |  1.666667  |
    >>> # |             minus_one  |    mse     |  0.666667  |  <-
    >>> # ----------------------------------------------------------
    >>> # ==========================================================
    >>> # |             identical  |    mae     |  1.000000  |
    >>> # |             minus_one  |    mae     |  0.666667  |  <-
    >>> # ----------------------------------------------------------
    >>> comparer = Comparer(patterns, estimators).compare(x, y)
    >>> # {'mse': {'identical': 1.666667, 'minus_one': 0.666666},
    >>> # 'mae': {'identical': 1.0, 'minus_one': 0.666666}}
    >>> print(comparer.scores)
    >>> # {'mse': 'minus_one', 'mae': 'minus_one'}
    >>> print(comparer.best_methods)

    """

    def __init__(self,
                 model_patterns: Dict[str, Union[ModelPattern, Dict[str, ModelPattern]]],
                 estimators: List[Estimator]):
        self.model_patterns = model_patterns
        self.estimators = dict(zip([estimator.type for estimator in estimators], estimators))

    @property
    def scores(self) -> Dict[str, Dict[str, Union[float, None]]]:
        return {k: v.scores for k, v in self.estimators.items()}

    @property
    def best_methods(self) -> Dict[str, str]:
        return {k: v.best_method for k, v in self.estimators.items()}

    def compare(self,
                x: np.ndarray,
                y: np.ndarray,
                *,
                verbose: bool = True) -> "Comparer":
        for estimator in self.estimators.values():
            methods = {}
            for model_name, pattern in self.model_patterns.items():
                if isinstance(pattern, dict):
                    pattern = pattern.get(estimator.type)
                if pattern is None:
                    continue
                methods[model_name] = pattern.predict
            estimator.estimate(x, y, methods, verbose=verbose)
        return self


class ScalarEMA:
    """
    Util class to record Exponential Moving Average (EMA) for scalar value

    Parameters
    ----------
    decay : float, decay rate for EMA
        * Formula: new = (1 - decay) * current + decay * history; history = new

    Examples
    --------
    >>> ema = ScalarEMA(0.5)
    >>> for i in range(4):
    >>>     print(ema.update("score", 0.5 ** i))  # 1, 0.75, 0.5, 0.3125

    """

    def __init__(self, decay):
        self._decay = decay
        self._ema_records = {}

    def __str__(self):
        return f"ScalarEMA({self._decay})"

    __repr__ = __str__

    def get(self, name):
        return self._ema_records.get(name)

    def update(self, name, new_value):
        history = self._ema_records.get(name)
        if history is None:
            updated = new_value
        else:
            updated = (1 - self._decay) * new_value + self._decay * history
        self._ema_records[name] = updated
        return updated


class Grid:
    """
    Util class provides permutation of simple, flattened param dicts
    * For permutation of complex, nested param dicts, please refers to `ParamGenerator` in `cftool.param_utils.basis`

    Parameters
    ----------
    param_grid : dict[str, list(int)]
        Indicates param names and corresponding possible values

    Examples
    ----------
    >>> grid = Grid({"a": [1, 2, 3], "b": [1, 2, 3]})
    >>> for param in grid:
    >>>     print(param)
    >>> # output : {'a': 1, 'b': 1}, {'a': 1, 'b': 2}, {'a': 1, 'b': 3}
    >>> #          {'a': 2, 'b': 1}, ..., {'a': 3, 'b': 3}

    """

    def __init__(self, param_grid):
        self._grid = param_grid

    def __iter__(self):
        items = sorted(self._grid.items())
        if not items:
            yield {}
        else:
            keys, values = zip(*items)
            for v in map(list, product(*values)):
                for i, vv in enumerate(v):
                    if isinstance(vv, tuple) and len(vv) == 1:
                        v[i] = vv[0]
                params = dict(zip(keys, v))
                yield params


class Visualizer:
    """
    Visualization class

    Methods
    ----------
    bar(self, data, classes, categories, save_name="bar_plot", title="",
            padding=1e-3, expand_floor=5, replace=True)
        Make bar plot with given `data`
        * data : np.ndarray
            Containing values for the bar plot, where data.shape =
            * (len(categories), ), if len(classes) == 1
            * (len(classes), len(categories)), otherwise
        * classes : list(str), list of str which indicates each class
            * each class will has its own color
            * len(classes) indicates how many bars are there in one category (side by side)
        * categories : list(str), list of str which indicates each category
            * a category will be a tick along x-axis
        * save_name : str, saving name of this bar plot
        * title : str, title of this bar plot
        * padding : float, minimum value of each bar
        * expand_floor : int, when len(categories) > `expand_floor`, the width of the figure will expand
            * for len(classes) == 1, `expand_floor` will be multiplied by 2 internally
        * overwrite : bool
            whether overwrite the existing file with the same file name of this plot's saving name

    function(self, f, x_min, x_max, classes, categories, save_names=None,
             n_sample=1000, expand_floor=5, overwrite=True):
        Make multiple (len(categories)) line plots with given function (`f`)
        * f : function
            * input should be an np.ndarray with shape == (n, n_categories)
            * output should be an np.ndarray with shape == (n, n_categories, n_categories)
        * x_min : np.ndarray, minimum x-values for each line plot
            * len(x_min) should be len(categories)
        * x_max : np.ndarray, maximum x-values for each line plot
            * len(x_max) should be len(categories)
        * classes : list(str), list of str which indicates each class
            * each class will has its own color
            * len(classes) indicates how many bars are there in one category (side by side)
        * categories : list(str), list of str which indicates each category
            * every category will correspond to a line plot
        * save_names : list(str), saving names of these line plots
        * n_sample : int, sample density along x-axis
        * expand_floor : int, the width of the figures will be expanded with ratios calculated by:
            expand_ratios = np.maximum(1., np.abs(x_min) / expand_floor, x_max / expand_floor)
        * overwrite : bool
            whether overwrite the existing file with the same file name of this plot's saving name

    """

    def __init__(self, export_folder):
        self.export_folder = os.path.abspath(export_folder)
        os.makedirs(self.export_folder, exist_ok=True)

    def _get_save_name(self, save_name):
        counter, tmp_save_name = 0, save_name
        while os.path.isfile(os.path.join(self.export_folder, f"{tmp_save_name}.png")):
            counter += 1
            tmp_save_name = f"{save_name}_{counter}"
        return tmp_save_name

    def bar(self, data: np.ndarray, classes: list, categories: list, save_name="bar_plot", title="",
            padding=1e-3, expand_floor=5, overwrite=True):
        n_class, n_categories = map(len, [classes, categories])
        data = [data / data.sum() + padding] if n_class == 1 else data / data.sum(0) + padding
        expand_floor = expand_floor * 2 if n_class == 1 else expand_floor
        colors = plt.cm.Paired([i / n_class for i in range(n_class)])
        x_base = np.arange(1, n_categories + 1)
        expand_ratio = max(1., n_categories / expand_floor)
        fig = plt.figure(figsize=(6.4 * expand_ratio, 4.8))
        plt.title(title)
        n_divide = n_class - 1
        width = 0.35 / max(1, n_divide)
        cls_ratio = 0.5 if n_class == 1 else 1
        for cls in range(n_class):
            plt.bar(x_base - width * (0.5 * n_divide - cls_ratio * cls), data[cls], width=width,
                    facecolor=colors[cls], edgecolor="white",
                    label=classes[cls])
        plt.xticks(
            [i for i in range(len(categories) + 2)],
            [""] + categories + [""]
        )
        plt.legend()
        plt.setp(plt.xticks()[1], rotation=30, horizontalalignment='right')
        plt.ylim(0, 1.2 + padding)
        fig.tight_layout()
        if not overwrite:
            save_name = self._get_save_name(save_name)
        plt.savefig(os.path.join(self.export_folder, f"{save_name}.png"))
        plt.close()

    def function(self, f: Callable[[np.ndarray], np.ndarray], x_min: np.ndarray, x_max: np.ndarray,
                 classes: list, categories: list, save_names=None, n_sample=1000,
                 expand_floor=5, overwrite=True):
        n_class, n_categories = map(len, [classes, categories])
        gaps = x_max - x_min
        x_base = np.linspace(x_min - 0.1 * gaps, x_max + 0.1 * gaps, n_sample)
        f_values = np.split(f(x_base), n_class, axis=1)
        if save_names is None:
            save_names = ["function_plot"] * n_categories
        colors = plt.cm.Paired([i / n_class for i in range(n_class)])
        expand_ratios = np.maximum(1., np.abs(x_min) / expand_floor, x_max / expand_floor)
        for i, (category, save_name, ratio, local_min, local_max, gap) in enumerate(zip(
                categories, save_names, expand_ratios, x_min, x_max, gaps)):
            plt.figure(figsize=(6.4 * ratio, 4.8))
            plt.title(f"pdf for {category}")
            local_base = x_base[..., i]
            for c in range(n_class):
                f_value = f_values[c][..., i].ravel()
                plt.plot(local_base, f_value, c=colors[c], label=f"class: {classes[c]}")
            plt.xlim(local_min - 0.2 * gap, local_max + 0.2 * gap)
            plt.legend()
            if not overwrite:
                save_name = self._get_save_name(save_name)
            plt.savefig(os.path.join(self.export_folder, f"{save_name}.png"))

    @staticmethod
    def visualize1d(method: Callable,
                    x: np.ndarray,
                    y: np.ndarray = None,
                    *,
                    title: str = None,
                    num_samples: int = 100,
                    expand_ratio: float = 0.25,
                    return_canvas: bool = False) -> Union[None, np.ndarray]:
        if x.shape[1] != 1:
            raise ValueError("visualize1d only supports 1-dimensional features")
        plt.figure()
        plt.title(title)
        if y is not None:
            plt.scatter(x, y, c="g", s=20)
        x_min, x_max = x.min(), x.max()
        expand = expand_ratio * (x_max - x_min)
        x0 = np.linspace(x_min - expand, x_max + expand, num_samples).reshape([-1, 1])
        plt.plot(x0, method(x0).ravel())
        return show_or_return(return_canvas)

    @staticmethod
    def visualize2d(method,
                    x: np.ndarray,
                    y: np.ndarray = None,
                    *,
                    title: str = None,
                    dense: int = 200,
                    padding: float = 0.1,
                    return_canvas: bool = False,
                    draw_background: bool = True,
                    extra_scatters: np.ndarray = None,
                    emphasize_indices: np.ndarray = None) -> Union[None, np.ndarray]:
        axis = x.T
        if axis.shape[0] != 2:
            raise ValueError("visualize2d only supports 2-dimensional features")
        nx, ny, padding = dense, dense, padding
        x_min, x_max = np.min(axis[0]), np.max(axis[0])  # type: float
        y_min, y_max = np.min(axis[1]), np.max(axis[1])  # type: float
        x_padding = max(abs(x_min), abs(x_max)) * padding
        y_padding = max(abs(y_min), abs(y_max)) * padding
        x_min -= x_padding
        x_max += x_padding
        y_min -= y_padding
        y_max += y_padding

        def get_base(_nx, _ny):
            _xf = np.linspace(x_min, x_max, _nx)
            _yf = np.linspace(y_min, y_max, _ny)
            n_xf, n_yf = np.meshgrid(_xf, _yf)
            return _xf, _yf, np.c_[n_xf.ravel(), n_yf.ravel()]

        xf, yf, base_matrix = get_base(nx, ny)
        z = method(base_matrix).reshape((nx, ny))

        labels = y.ravel()
        num_labels = y.max().item() + 1
        colors = plt.cm.rainbow([i / num_labels for i in range(num_labels)])[labels]

        plt.figure()
        plt.title(title)
        if draw_background:
            xy_xf, xy_yf = np.meshgrid(xf, yf, sparse=True)
            plt.pcolormesh(xy_xf, xy_yf, z, cmap=plt.cm.Pastel1)
        else:
            plt.contour(xf, yf, z, c='k-', levels=[0])
        plt.scatter(axis[0], axis[1], c=colors)

        if emphasize_indices is not None:
            indices = np.array([False] * len(axis[0]))
            indices[np.asarray(emphasize_indices)] = True
            plt.scatter(axis[0][indices], axis[1][indices], s=80,
                        facecolors="None", zorder=10)
        if extra_scatters is not None:
            plt.scatter(*np.asarray(extra_scatters).T, s=80, zorder=25, facecolors="red")

        return show_or_return(return_canvas)


__all__ = [
    "Anneal", "Metrics", "ScalarEMA", "Grid", "Visualizer",
    "Estimator", "ModelPattern", "Comparer"
]
